"""Friendlier APIs for Trello."""
import re
from abc import ABC
from dataclasses import dataclass, field
from functools import cached_property as property
from typing import Any, Mapping, Optional, Sequence, Union, cast

import trello


@dataclass
class Comment:
    """Friendlier API for a Trello comment."""

    obj: Mapping[str, Any]

    @property
    def member(self) -> str:
        return cast(str, self.obj["memberCreator"]["username"])

    @property
    def date(self) -> str:
        return cast(str, self.obj["date"][:10])

    @property
    def body(self) -> str:
        return cast(str, self.obj["data"]["text"])


@dataclass
class Attachment:
    """Friendlier API for a Trello attachment."""

    obj: Mapping[str, Any]

    @property
    def url(self) -> str:
        return cast(str, self.obj["url"])

    @property
    def name(self) -> str:
        return cast(str, self.obj["name"])


@dataclass
class ChecklistItem:
    """Friendlier API for a Trello checklist item."""

    obj: Mapping[str, Any]

    @property
    def name(self) -> str:
        return cast(str, self.obj["name"])

    @property
    def checked(self) -> bool:
        return cast(bool, self.obj["checked"])


@dataclass
class Checklist:
    """Friendlier API for a Trello checklist."""

    obj: trello.Checklist

    @property
    def name(self) -> str:
        return cast(str, self.obj.name)

    @property
    def items(self) -> Sequence[ChecklistItem]:
        return [ChecklistItem(x) for x in self.obj.items]


class Exportable(ABC):
    """Base class for an exportable Trello object."""

    obj: Any

    @property
    def url(self) -> str:
        return cast(str, self.obj.url)

    @property
    def slug(self) -> str:
        return self.url.split("/")[-1]


@dataclass
class Card(Exportable):
    """Friendlier API for a Trello card."""

    obj: trello.Card

    @property
    def name(self) -> str:
        return cast(str, self.obj.name)

    @property
    def description(self) -> str:
        return cast(str, self.obj.description)

    @property
    def due_date(self) -> Optional[str]:
        return self.obj.due[:10] if self.obj.due else None

    @property
    def labels(self) -> Sequence[str]:
        return [x.name if x.name else x.color for x in (self.obj.labels or [])]

    @property
    def members(self) -> Sequence[str]:
        return [
            self.obj.board.client.get_member(x).username for x in self.obj.member_ids
        ]

    @property
    def checklists(self) -> Sequence[Checklist]:
        return [Checklist(x) for x in self.obj.checklists]

    @property
    def attachments(self) -> Sequence[Attachment]:
        return [Attachment(x) for x in self.obj.attachments]

    @property
    def comments(self) -> Sequence[Comment]:
        return [Comment(x) for x in self.obj.comments]


@dataclass
class CardList:
    """Friendlier API for a Trello list.

    Not named `List` because that clashes with `typing.List`.
    """

    obj: trello.List

    @property
    def name(self) -> str:
        return cast(str, self.obj.name)

    @property
    def cards(self) -> Sequence[Card]:
        return [Card(x) for x in self.obj.list_cards()]


@dataclass
class Board(Exportable):
    """Friendlier API for a Trello board."""

    obj: trello.Board

    @property
    def name(self) -> str:
        return cast(str, self.obj.name)

    @property
    def lists(self) -> Sequence[CardList]:
        return [CardList(x) for x in self.obj.open_lists()]


Config = Mapping[str, str]


@dataclass
class Client:
    """Friendlier API for the Trello client."""

    credentials: Config
    obj: trello.TrelloClient = field(init=False)

    def __post_init__(self) -> None:
        self.obj = trello.TrelloClient(**self.credentials)

    def get_url(self, url: str) -> Union[Board, Card]:
        """Return a Trello board or card from a URL."""
        url_match = re.search(r"trello\.com/([bc])/([\w]*)", url)
        if not url_match:
            raise ValueError(f"Invalid URL: {url}")

        object_type, object_id = url_match.group(1, 2)

        if object_type == "b":
            return Board(self.obj.get_board(object_id))

        return Card(self.obj.get_card(object_id))


def get_credentials(app_name: str) -> Config:
    """Prompt for and generate API credentials."""
    print("Go to the following link in your browser:")
    print("https://trello.com/app-key")
    print()

    api_key = input("Enter your API key: ")
    api_secret = input("Enter your API secret: ")
    print()

    access_token = trello.util.create_oauth_token(
        expiration="never",
        scope="read",
        key=api_key,
        secret=api_secret,
        name=app_name,
        output=False,
    )

    return {
        "api_key": api_key,
        "token": access_token["oauth_token"],
    }



import sys
from io import StringIO

from trello2md import api


def test_get_credentials(monkeypatch):
    inputs = StringIO(
        # Enter your API key:
        "api_key\n"
        # Enter your API secret:
        "api_secret\n"
    )
    monkeypatch.setattr(sys, "stdin", inputs)

    def stub_create_token(**kwargs):
        assert kwargs == {
            "expiration": "never",
            "scope": "read",
            "key": "api_key",
            "secret": "api_secret",
            "name": "trello2md",
            "output": False,
        }
        return {"oauth_token": "oauth_token"}

    monkeypatch.setattr(api.trello.util, "create_oauth_token", stub_create_token)

    credentials = api.get_credentials("trello2md")

    assert credentials == {"api_key": "api_key", "token": "oauth_token"}



import filecmp
import runpy
import stat
import subprocess
import sys
from importlib.metadata import entry_points
from pathlib import Path

import pytest

from trello2md import cli

TESTS_DIRNAME = Path(__file__).parent.resolve()
COMMAND_NAME = __name__

pytestmark = [
    pytest.mark.vcr(filter_headers=["authorization"]),
    pytest.mark.block_network,
]


# Autouse to ensure local config file isn't used by tests
@pytest.fixture(autouse=True)
def config_path(tmp_path, monkeypatch):
    config_path = tmp_path / "config"
    monkeypatch.setattr(cli, "CONFIG_PATH", config_path)
    return config_path


def test_write_credentials(config_path, monkeypatch, capsys):
    monkeypatch.setattr(sys, "argv", [COMMAND_NAME, "auth"])

    def stub_get_credentials(app_name):
        assert app_name == "trello2md"
        return {"api_key": "api_key", "token": "oauth_token"}

    monkeypatch.setattr(cli.api, "get_credentials", stub_get_credentials)

    cli.main()

    assert config_path.read_text() == (
        "[credentials]\napi_key = api_key\ntoken = oauth_token\n\n"
    )

    assert stat.S_IMODE(config_path.stat().st_mode) == stat.S_IRUSR | stat.S_IWUSR

    captured = capsys.readouterr()
    assert str(config_path) in captured.out


@pytest.fixture
def config_file(config_path, monkeypatch):
    # Assuming sample board is public
    config_path.write_text("[credentials]\napi_key = \ntoken = \n\n")
    return config_path


def test_write_board(config_file, tmp_path, monkeypatch, capsys):
    url = "https://trello.com/b/WODq2cwg/sample-board"
    md_dirname = "sample-board"

    monkeypatch.chdir(tmp_path)
    monkeypatch.setattr(sys, "argv", [COMMAND_NAME, url])

    cli.main()

    captured = capsys.readouterr()

    md_filenames = [
        "index.md",
        "sample-card.md",
        "copied-card.md",
        "another-card.md",
        "another-card-1.md",
        "another-card-2.md",
    ]

    for filename, output in zip(md_filenames, captured.out.splitlines()):
        assert output.endswith(filename)

    match, mismatch, errors = filecmp.cmpfiles(
        tmp_path / md_dirname, TESTS_DIRNAME / md_dirname, md_filenames, shallow=False
    )
    assert set(match) == set(md_filenames)
    assert not (mismatch or errors)


@pytest.mark.parametrize(
    "url, md_filename",
    [
        pytest.param(
            "https://trello.com/c/HGYGb5iM/2-sample-card", "sample-card.md", id="sample"
        ),
        pytest.param("https://trello.com/c/9lhLVJz3", "copied-card.md", id="copied"),
    ],
)
def test_write_card(url, md_filename, config_file, monkeypatch, capsys):
    md_dirname = "sample-board"

    monkeypatch.setattr(sys, "argv", [COMMAND_NAME, url])

    cli.main()

    with open(TESTS_DIRNAME / md_dirname / md_filename) as f:
        sample_card = f.read()

    captured = capsys.readouterr()
    assert captured.out == sample_card


def test_help():
    # Using subprocess for simplicity and to smoke-test command line use
    result = subprocess.run(["trello2md", "-h"], check=True, capture_output=True)
    assert result.stdout.startswith(b"Export Trello boards and cards to Markdown.")


def test_no_arguments(config_file):
    # Using subprocess for simplicity and to smoke-test command line use
    result = subprocess.run(["trello2md"], capture_output=True)
    assert result.returncode > 0
    assert result.stderr.startswith(b"Usage")


def test_invalid_argument(config_file, monkeypatch):
    url = "https://trello.com/b"

    # Not using subprocess because it doesn't measure coverage (by default)

    monkeypatch.setattr(sys, "argv", [COMMAND_NAME, url])

    with pytest.raises(ValueError, match=url):
        cli.main()


def test_console_script(monkeypatch):
    # This is overkill with the subprocess tests, but it gives 100% branch coverage
    # of __main__.py, and it's an interesting example of using importlib

    # Don't leak loaded module into other tests (esp. test_run_as_module)
    monkeypatch.setattr(sys, "modules", dict(sys.modules))

    entry = next(
        x for x in entry_points()["console_scripts"] if x.name == "trello2md"
    )  # pragma: no cover

    console_script = entry.load()

    assert console_script == cli.main


def test_run_as_module(monkeypatch):
    monkeypatch.setattr(cli, "main", lambda: 1 / 0)

    with pytest.raises(ZeroDivisionError):
        runpy.run_module("trello2md", run_name="__main__")



"""Export Trello boards and cards to Markdown.

Usage:
  trello2md auth
  trello2md <url>
  trello2md -h | --help | --version

Arguments:
  auth      Authorize use of the Trello API
  <url>     The URL for a Trello board or card

"""
import builtins
import os
import re
import stat
import sys
from configparser import ConfigParser
from contextlib import contextmanager
from functools import partial
from importlib import metadata
from pathlib import Path
from typing import IO, Iterator, Optional

from docopt import docopt

from . import api

COMMAND_NAME = "trello2md"
CONFIG_PATH = Path.home() / ".config" / COMMAND_NAME


def main() -> Optional[str]:
    """Process command line arguments."""
    args = docopt(__doc__, version=metadata.version(COMMAND_NAME))

    if args["auth"]:
        write_credentials()
    else:
        write_url(args["<url>"])

    return None


def write_credentials() -> None:
    """Prompt for and save API credentials."""
    config = ConfigParser()
    config["credentials"] = api.get_credentials(COMMAND_NAME)

    with open(CONFIG_PATH, "w") as config_file:
        config.write(config_file)

    CONFIG_PATH.chmod(stat.S_IRUSR | stat.S_IWUSR)

    print()
    print(f"Credentials saved to {CONFIG_PATH}")


def write_url(url: str) -> None:
    """Print Markdown from a Trello URL."""
    config = ConfigParser()

    with open(CONFIG_PATH) as config_file:
        config.read_file(config_file)

    client = api.Client(config["credentials"])

    obj = client.get_url(url)

    if isinstance(obj, api.Board):
        with open_board(obj) as file:
            write_board(obj, file)
    else:
        write_card(obj)


@contextmanager
def open_board(board: api.Board) -> Iterator[IO[str]]:
    """Yield an open file in a new directory for a Trello board."""
    dirname = get_filename(board.slug)
    os.mkdir(dirname)
    os.chdir(dirname)

    filename = "index.md"
    print(f"{board.name:30.30} -> {os.path.join(dirname, filename)}")
    with open(filename, "w") as file:
        yield file


def write_board(board: api.Board, file: Optional[IO[str]] = None) -> None:
    """Print Markdown for a Trello board."""
    print = partial(builtins.print, file=file if file else sys.stdout)

    print(f"# {board.name}")

    for lst in board.lists:
        print(f"\n## {lst.name}")

        if lst.cards:
            print()

        for card in lst.cards:
            with open_card(card) as file:
                write_card(card, file=file)

                print(f"- [{card.name}]({file.name})", end="")
                print(f" {meta}" if (meta := get_card_meta(card)) else "")


@contextmanager
def open_card(card: api.Card) -> Iterator[IO[str]]:
    """Yield an open file for a Trello card."""
    filename = get_filename(card.slug, ".md")
    print(f"{card.name:30.30} -> {filename}")
    with open(filename, "w") as file:
        yield file


def write_card(card: api.Card, file: Optional[IO[str]] = None) -> None:
    """Print Markdown for a Trello card."""
    print = partial(builtins.print, file=file if file else sys.stdout)

    print(f"# {card.name}")

    if meta := get_card_meta(card):
        print()
        print(meta)

    if card.description:
        print(f"\n{card.description}")

    for checklist in card.checklists:
        print(f"\n## {checklist.name}")

        if checklist.items:
            print()

        for item in checklist.items:
            print(f"- [{'x' if item.checked else ' '}] {item.name}")

    if card.attachments:
        print("\n## Attachments")
        print()

    for attachment in card.attachments:
        if attachment.name == attachment.url:
            print(f"- <{attachment.url}>")
        else:
            print(f"- [{attachment.name}]({attachment.url})")

    if card.comments:
        print("\n## Comments")

    for comment in card.comments:
        print(f"\n### {comment.member} on {comment.date}\n")
        print(comment.body)


def get_filename(slug: str, ext: str = "") -> str:
    """Return a unique filename."""
    # Remove numeric prefix from card slug
    slug = re.sub(r"^\d+-", "", slug)

    filename = slug + ext
    suffix = 0

    while os.path.exists(filename):
        filename = f"{slug}-{(suffix := suffix + 1)}{ext}"

    return filename


def get_card_meta(card: api.Card) -> str:
    """Return metadata for a Trello card, separated by commas."""
    members = [f"@{x}" for x in card.members]
    labels = [f"`{x}`" for x in card.labels]
    return ", ".join([x for x in [card.due_date, *members, *labels] if x])
